@@ -152,7 +152,8 @@ class WatWriter : ModuleContext {
   void WriteDataSegment(const DataSegment& segment);
   void WriteImport(const Import& import);
   void WriteExport(const Export& export_);
-  void WriteFuncType(const FuncType& func_type);
+  void WriteTypeEntry(const TypeEntry& type);
+  void WriteField(const Field& field);
   void WriteStartFunction(const Var& start);
 
   class ExprVisitorDelegate;
@@ -184,7 +185,7 @@ class WatWriter : ModuleContext {
   Index global_index_ = 0;
   Index table_index_ = 0;
   Index memory_index_ = 0;
-  Index func_type_index_ = 0;
+  Index type_index_ = 0;
   Index event_index_ = 0;
   Index data_segment_index_ = 0;
   Index elem_segment_index_ = 0;
@@ -377,7 +378,7 @@ void WatWriter::WriteBrVar(const Var& var, NextChar next_char) {
 }
 
 void WatWriter::WriteType(Type type, NextChar next_char) {
-  const char* type_name = GetTypeName(type);
+  const char* type_name = type.GetName();
   assert(type_name);
   WritePuts(type_name, next_char);
 }
@@ -426,27 +427,25 @@ void WatWriter::WriteEndBlock() {
 }
 
 void WatWriter::WriteConst(const Const& const_) {
-  switch (const_.type) {
+  switch (const_.type()) {
     case Type::I32:
       WritePutsSpace(Opcode::I32Const_Opcode.GetName());
-      Writef("%d", static_cast<int32_t>(const_.u32));
+      Writef("%d", static_cast<int32_t>(const_.u32()));
       WriteNewline(NO_FORCE_NEWLINE);
       break;
 
     case Type::I64:
       WritePutsSpace(Opcode::I64Const_Opcode.GetName());
-      Writef("%" PRId64, static_cast<int64_t>(const_.u64));
+      Writef("%" PRId64, static_cast<int64_t>(const_.u64()));
       WriteNewline(NO_FORCE_NEWLINE);
       break;
 
     case Type::F32: {
       WritePutsSpace(Opcode::F32Const_Opcode.GetName());
       char buffer[128];
-      WriteFloatHex(buffer, 128, const_.f32_bits);
+      WriteFloatHex(buffer, 128, const_.f32_bits());
       WritePutsSpace(buffer);
-      float f32;
-      memcpy(&f32, &const_.f32_bits, sizeof(f32));
-      Writef("(;=%g;)", f32);
+      Writef("(;=%g;)", Bitcast<float>(const_.f32_bits()));
       WriteNewline(NO_FORCE_NEWLINE);
       break;
     }
@@ -454,20 +453,18 @@ void WatWriter::WriteConst(const Const& const_) {
     case Type::F64: {
       WritePutsSpace(Opcode::F64Const_Opcode.GetName());
       char buffer[128];
-      WriteDoubleHex(buffer, 128, const_.f64_bits);
+      WriteDoubleHex(buffer, 128, const_.f64_bits());
       WritePutsSpace(buffer);
-      double f64;
-      memcpy(&f64, &const_.f64_bits, sizeof(f64));
-      Writef("(;=%g;)", f64);
+      Writef("(;=%g;)", Bitcast<double>(const_.f64_bits()));
       WriteNewline(NO_FORCE_NEWLINE);
       break;
     }
 
     case Type::V128: {
       WritePutsSpace(Opcode::V128Const_Opcode.GetName());
-      Writef("i32x4 0x%08x 0x%08x 0x%08x 0x%08x", const_.vec128.v[0],
-             const_.vec128.v[1], const_.vec128.v[2],
-             const_.vec128.v[3]);
+      auto vec = const_.vec128();
+      Writef("i32x4 0x%08x 0x%08x 0x%08x 0x%08x", vec.u32(0), vec.u32(1),
+             vec.u32(2), vec.u32(3));
       WriteNewline(NO_FORCE_NEWLINE);
       break;
     }
@@ -550,6 +547,7 @@ class WatWriter::ExprVisitorDelegate : public ExprVisitor::Delegate {
   Result OnThrowExpr(ThrowExpr*) override;
   Result OnRethrowExpr(RethrowExpr*) override;
   Result OnAtomicWaitExpr(AtomicWaitExpr*) override;
+  Result OnAtomicFenceExpr(AtomicFenceExpr*) override;
   Result OnAtomicNotifyExpr(AtomicNotifyExpr*) override;
   Result OnAtomicLoadExpr(AtomicLoadExpr*) override;
   Result OnAtomicStoreExpr(AtomicStoreExpr*) override;
@@ -890,6 +888,13 @@ Result WatWriter::ExprVisitorDelegate::OnAtomicWaitExpr(AtomicWaitExpr* expr) {
   return Result::Ok;
 }
 
+Result WatWriter::ExprVisitorDelegate::OnAtomicFenceExpr(
+    AtomicFenceExpr* expr) {
+  assert(expr->consistency_model == 0);
+  writer_->WritePutsNewline(Opcode::AtomicFence_Opcode.GetName());
+  return Result::Ok;
+}
+
 Result WatWriter::ExprVisitorDelegate::OnAtomicNotifyExpr(
     AtomicNotifyExpr* expr) {
   writer_->WriteLoadStoreExpr<AtomicNotifyExpr>(expr);
@@ -926,7 +931,7 @@ Result WatWriter::ExprVisitorDelegate::OnTernaryExpr(TernaryExpr* expr) {
 Result WatWriter::ExprVisitorDelegate::OnSimdLaneOpExpr(SimdLaneOpExpr* expr) {
   writer_->WritePutsSpace(expr->opcode.GetName());
   writer_->Writef("%" PRIu64, (expr->val));
-  writer_->WritePutsNewline("");
+  writer_->WriteNewline(NO_FORCE_NEWLINE);
   return Result::Ok;
 }
 
@@ -935,9 +940,9 @@ Result WatWriter::ExprVisitorDelegate::OnSimdShuffleOpExpr(
   writer_->WritePutsSpace(expr->opcode.GetName());
   std::array<uint8_t, 16> values = Bitcast<std::array<uint8_t, 16>>(expr->val);
   for (int32_t lane = 0; lane < 16; ++lane) {
-    writer_->Writef(" %u", values[lane]);
+    writer_->Writef("%u", values[lane]);
   }
-  writer_->WritePutsNewline("");
+  writer_->WriteNewline(NO_FORCE_NEWLINE);
   return Result::Ok;
 }
 
@@ -1264,6 +1269,10 @@ void WatWriter::WriteElemSegment(const ElemSegment& segment) {
     WriteInitExpr(segment.offset);
   }
 
+  if (flags == SegDeclared) {
+    WritePuts("declare", NextChar::Space);
+  }
+
   if (flags & SegUseElemExprs) {
     WriteType(segment.elem_type, NextChar::Space);
   } else {
@@ -1361,15 +1370,52 @@ void WatWriter::WriteExport(const Export& export_) {
   WriteCloseNewline();
 }
 
-void WatWriter::WriteFuncType(const FuncType& func_type) {
+void WatWriter::WriteTypeEntry(const TypeEntry& type) {
   WriteOpenSpace("type");
-  WriteNameOrIndex(func_type.name, func_type_index_++, NextChar::Space);
-  WriteOpenSpace("func");
-  WriteFuncSigSpace(func_type.sig);
-  WriteCloseSpace();
+  WriteNameOrIndex(type.name, type_index_++, NextChar::Space);
+  switch (type.kind()) {
+    case TypeEntryKind::Func:
+      WriteOpenSpace("func");
+      WriteFuncSigSpace(cast<FuncType>(&type)->sig);
+      WriteCloseSpace();
+      break;
+
+    case TypeEntryKind::Struct: {
+      auto* struct_type = cast<StructType>(&type);
+      WriteOpenSpace("struct");
+      Index field_index = 0;
+      for (auto&& field : struct_type->fields) {
+        // TODO: Write shorthand if there is no name.
+        WriteOpenSpace("field");
+        WriteNameOrIndex(field.name, field_index++, NextChar::Space);
+        WriteField(field);
+        WriteCloseSpace();
+      }
+      WriteCloseSpace();
+      break;
+    }
+
+    case TypeEntryKind::Array: {
+      auto* array_type = cast<ArrayType>(&type);
+      WriteOpenSpace("array");
+      WriteField(array_type->field);
+      WriteCloseSpace();
+      break;
+    }
+  }
   WriteCloseNewline();
 }
 
+void WatWriter::WriteField(const Field& field) {
+  if (field.mutable_) {
+    WriteOpenSpace("mut");
+  }
+  WriteType(field.type, NextChar::Space);
+  if (field.mutable_) {
+    WriteCloseSpace();
+  }
+}
+
 void WatWriter::WriteStartFunction(const Var& start) {
   WriteOpenSpace("start");
   WriteVar(start, NextChar::None);
@@ -1414,8 +1460,8 @@ Result WatWriter::WriteModule() {
       case ModuleFieldType::DataSegment:
         WriteDataSegment(cast<DataSegmentModuleField>(&field)->data_segment);
         break;
-      case ModuleFieldType::FuncType:
-        WriteFuncType(cast<FuncTypeModuleField>(&field)->func_type);
+      case ModuleFieldType::Type:
+        WriteTypeEntry(*cast<TypeModuleField>(&field)->type);
         break;
       case ModuleFieldType::Start:
         WriteStartFunction(cast<StartModuleField>(&field)->start);
